unit Unit1;

{
  WinThrottle source code
  Written in Delphi 7 Enterprise
  Should work on all versions of Delphi 7 and 'may' compile on lower versions
  No external Delphi components are needed. Compiles with just Delphi 7 'out of the box'
  No external Delphi DLLs or other Delphi modules needed.
  No need to install the Windows way. Just run the executable.
  Needs 'GIVEIO.SYS', a public domain driver which should be copied to your Windows\System32 directory.

  Based upon the original WinThrottle by Jeff Leyda.

  The author hereby declares that no copyright exists on any part of the code.
  The WinThrottle core was originally written by Jeff Leyda.
  The Delphi code was mostly written by Bill Jameson.
  Parts of the code (registry, histogram) were taken from various Delphi
  source code sites. Any external source code used is Public Domain code
  without any mention of the author(s).

  The author of this program hereby donates the source code to the public domain.

  This is a working but still unfinished project. Feel free to modify it any way
  you like. Please donate your source code to others and especially: please send
  your working source code project back to Jeff Leyda so that he may add your
  project to his WinThrottle Web site.

  Many thanks to Jeff Leyda for his original source code, his feedback & helpful
  explanations.

  To Do:

  - The code which reads the CPU temperature was explicitly written for
  the (most commonly used) LM sensor. Other sensors may require modification
  of the code. It is advised to leave the existing code in place and only 'add'
  new sensor types. If automatic detection doesn't work, give the user a radio
  button choice. They can experiment themselves to find a working sensor.

  - The Auto-Throttle works, but may need improvement. The slider wildly
  jumps back and forth, immmediately responding to any CPU temperature change.
  There already is a mechanism which 'averages' the last 5 reads. It works but
  somehow it doesn't 'feel' right.

  - Expand the 'Settings' tab with more choices to turn on/off the histogram
  and other rarely used functions. This may save a few processor cycles.

  - Add an option to turn WinThrottle into an automatic power saver. You need
  to add a Windows 'Permance Monitor' object to read the current CPU load. If
  the CPU load drops below a certain level, WinThrottle could throttle down
  the CPU to save power. Unfortunately, creating a Performance Monitor is
  well-documented in VB but not in C++ or Delphi. It involves a lot of
  programming and there is no 'standard' object available.
}




interface

uses
  Windows, Messages, SysUtils, Classes, Forms, Graphics,
  Dialogs, Registry, StdCtrls, StrUtils, ComCtrls, Controls, ExtCtrls, Mask,
  Spin, ShellAPI, Menus;

type
  TForm1 = class(TForm)
    Timer1: TTimer;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    Label1: TLabel;
    TrackBar1: TTrackBar;
    Button1: TButton;
    Panel1: TPanel;
    Label2: TLabel;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    TabSheet4: TTabSheet;
    Memo1: TMemo;
    Image1: TImage;
    CheckBox1: TCheckBox;
    Label4: TLabel;
    SpinEdit2: TSpinEdit;
    SpinEdit1: TSpinEdit;
    Label3: TLabel;
    Button2: TButton;
    CheckBox2: TCheckBox;
    Button3: TButton;
    Memo2: TMemo;
    procedure Button1Click(Sender: TObject);
    procedure EnumAllKeys(hkey: THandle);
    function GetRegistryData(RootKey: HKEY; Key, Value: string): variant;
    function openIO:Boolean;
    function GetThrottleSlider:Integer;
    procedure FormCreate(Sender: TObject);
    function DetectChipset:Boolean;
    procedure GetThrottleValFromChip;
    Procedure SetThrottleValue(Throttle:Byte);
    Procedure ProbeThrottleIncrement;
    procedure TrackBar1Change(Sender: TObject);
    Function ReadTemp:Word;
    procedure Timer1Timer(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
    Procedure WriteGraph;
    procedure ClearImage;
    procedure Button2Click(Sender: TObject);
    procedure AutoThrottle(CurrentTemp:Integer);
    procedure SetAutoThrottle(Value:Integer);
    procedure FormDestroy(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Restore1Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure SetRegistryData(RootKey: HKEY; Key, Value: string; RegDataType: TRegDataType; Data: variant);
    Procedure SaveSettings;
    procedure Button3Click(Sender: TObject);
    procedure CheckBox2MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure CheckBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure SpinEdit1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure SpinEdit2MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);

  private
    TrayIcon: TNotifyIconData;
    procedure RestoreMainForm;
    procedure MinimizeToTray(Sender: TObject);

    { Private declarations }
  public
    { Public declarations }
  end;

Const
  THROTTLE_ENABLE_BIT=$10;	// bit 4 is the enable bit for most (all?) ACPI software throttles.
  THROTTLE_REGISTER=$10;		// acpi base+10 is where to tweak our throttle register
  Scale=504; //504 second frame just fits into history window

var
  Form1: TForm1;
  StopFlag:Boolean;
  CurrentKeyName:String;
  CurrentValueName:String;
  RegBaseName:String;
  LimitKeyDepth: Integer;
  g_acpiBaseAddress: Word; //16-bit unsigned integer
  ThrottleValRead:Byte;
  Step625:Boolean;
  SliderText: Array[0..15] of AnsiString;
  ReadBackWorks:Boolean;
  TempArray : Array[0..Scale] of byte; //int64;
  TempIdx:integer;
  Counter:byte;
  Run:Array[0..7] of Boolean;
  AverageTemp:Array[0..4] of Integer;
  AverageTempIdx:Integer;

implementation

{$R *.dfm}

(*
At program start we try to find the ACPI control block.

If we fail, we simply bomb out since there's nothing else we can do.

If we succeed, the ACPI chip is probed.
(a) Is it a write-only chip which doesn't provide feedback?
(b) Does it support 6.25% step values?

Of course, there's no way of knowing if you hit a write-only
chip or just a wrong port without a chip attached. In this
case we simply continue 'blindly' and hope it works.

The throttle slider is preset to 6.25 step values but if the
6.25% test fails, the slider is set to 12.5% values by
adjusting the slider's 'max' property.

After that, we read the current throttle value from the chip
and pre-set the slider to it. For write-only chips we simply
preset the slider to 0 and ignore any readbacks.

Once the initialization phase is over, you can just call any
procedure without having to test anything. If you want, you
can set up a timer which calls the read/set routines or do
whatever else you like. To exit the program, just execute
an 'application.terminate' statement. Any value written into
the throttle register will remain there, even after program exit.

You could use this app as an automatic speed/temperature limiter.
If the CPU gets too hot you could automatically throttle down. But
if the CPU runs hot it's probably overloaded by another program. In
that case you will want to run this program at 'realtime' priority
so that it gets time to act before you get a CPU meltdown.

Change the priority class to whatever you like:

IDLE_PRIORITY_CLASS:      {lowest priority}
NORMAL_PRIORITY_CLASS:    {normal priority}
HIGH_PRIORITY_CLASS:      {serviced first}
REALTIME_PRIORITY_CLASS:  {time-critical task}
*)

procedure TForm1.FormCreate(Sender: TObject);
var
  i:integer;
  s:AnsiString;
  StartInSysTray:AnsiString;
begin
  Timer1.Enabled:=False;
	SetPriorityClass(GetCurrentProcess,REALTIME_PRIORITY_CLASS);
  AverageTempIdx:=0;
  SpinEdit1.Value:=50; // realistic auto-throttle start value
  for i:=0 to Scale do TempArray[i]:=0;
  TempIdx:=-1;
  ClearImage;
  LimitKeyDepth:=0;
  CurrentKeyName:='';
  CurrentValueName:='';
  RegBaseName:='';
  StopFlag:=False;
  g_acpiBaseAddress:=$CAFE;
  Step625:=False;
  ReadBackWorks:=False;
  Memo2.Clear;

  s:=GetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0', 'UseAutoThrottle');
  if s='1' then
  begin
    CheckBox1.Checked:=True;
    SpinEdit1.Enabled :=True;
    SpinEdit2.Enabled :=True;
    SpinEdit1.Color :=clWhite;
    SpinEdit2.Color :=clWhite;
    Trackbar1.Enabled:=False;
    Button1.Enabled:=False;
    for i:=0 to 7 do Run[i]:=False;
  end;

  StartInSysTray:=GetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0', 'StartInSysTray');

  s:=GetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0', 'AutoThrottleStart');
  if s<>'' then SpinEdit1.Value:=StrToInt(s);

  s:=GetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0', 'ThrottleRange');
  if s<>'' then SpinEdit2.Value:=StrToInt(s);

  Panel1.Caption:='';
  Label1.Caption:='';
  Label2.Caption:='Reading CPU';

  Memo1.Clear;
  Memo1.Lines.BeginUpdate;

  TrayIcon.cbSize := SizeOf(TrayIcon);
  TrayIcon.Wnd := Self.Handle;
  TrayIcon.uID := 0;
  TrayIcon.uFlags := NIF_ICON or NIF_TIP or NIF_MESSAGE;
  TrayIcon.uCallbackMessage := WM_MOUSEMOVE;
  TrayIcon.hIcon := Application.Icon.Handle;
  TrayIcon.szTip := 'Click-Left: Show   Click-Right: Exit';
  Application.OnMinimize := MinimizeToTray;

  SliderText[0]:='Off';
  SliderText[1]:='6.25%';
  SliderText[2]:='12.5%';
  SliderText[3]:='18.75%';
  SliderText[4]:='25%';
  SliderText[5]:='31.25%';
  SliderText[6]:='37.5%';
  SliderText[7]:='43.75%';
  SliderText[8]:='50%';
  SliderText[9]:='56.25%';
  SliderText[10]:='62.5%';
  SliderText[11]:='68.75%';
  SliderText[12]:='75%';
  SliderText[13]:='81.25%';
  SliderText[14]:='87.5%';
  SliderText[15]:='93.75%';

  if StartInSysTray='1' then
  begin
    Checkbox2.Checked:=True;
    Shell_notifyIcon(NIM_ADD, @TrayIcon);
  end
  else
  begin
    Checkbox2.Checked:=False;
    Self.Show;
  end;

  for i:=0 to 7 do Run[i]:=False;

  if openIO=False then
  begin
    showmessage('GIVEIO.SYS failed');
    application.Terminate;
  end;

  if DetectChipset=False then
    begin
    ShowMessage('Cannot find ACPI control block in registry');
    Application.Terminate;
  end;

  ProbeThrottleIncrement;
  GetThrottleValFromChip;
  Timer1.Enabled:=True;

  {the main routine just 'hangs' here without using}
  {any processor time, waiting for the slider to be}
  {moved or the button to be pushed.}
end;

(*
Clicking the 'SET' button is currently the only way to write a throttle
value into the register. After initialization you can just call the
SetThrottleValue procedure with a new throttle value. In this case the
'GetThrottleValFromChip' procedure is called to provide feedback. It reads
back the current throttle value from the chip and adjusts the slider.
If everything works as planned, the slider shouldn't move after you click 'SET'.
*)

procedure TForm1.Button1Click(Sender: TObject);
begin
  SetThrottleValue(GetThrottleSlider);
  GetThrottleValFromChip;
end;

(*
The ACPI control block should be at a fixed position in the registry.
However, evaluation versions of programs often insert bogus keys everywhere,
acting as 'hidden' keys for their time limits & as decoys for crackers.
Windows just ignores keys it doesn't need but for us, it makes it harder
to find the key we want.

So we start a seach for the ACPI control block from the HKLM\HARDWARE key.
If we find a key we try to open it and travel deeper into the registry.
At the HKLM\HARDWARE\ACPI\FADT level we enumerate all values and search
those values for the 'FACP' keyword.

If we get past the FADT key without finding the 'FACP' keyword, we simply
quit since there's nothing we can do.

Not finding the ACPI control block means that Windows doesn't have its
Power Management installed. The most likely cause is that the ACPI
power management in the machine's BIOS has been turned off.

Since it's a recursive search, we limit the search to 1000 keys to prevent
Delphi from blowing up the stack.

As the 'Current key' has an unwanted format, we maintain 'RegBaseName'
ourselves, which always contains the full path to the key.
In real life, it finds the key instantly.

The compiler issues a bogus warning about the 'v' TStringlist not having
being initialized. Just ignore the warning.

I donloaded this registry stuff from the Net (no copyright) and adapted
it to suit our needs.
*)

procedure TForm1.EnumAllKeys(hkey: THandle);
  var
    l: TStringList;
    v: TStringList;
    n: Integer;
    i,j,k: Integer;
    s,fadt: String;

  begin
    if StopFlag=True then exit;
    if LimitKeyDepth>1000 then exit;
    Inc(LimitKeyDepth);

    with TRegistry.Create do
      try
        RootKey := hkey;
        OpenKey(EmptyStr, False);
        l := TStringList.Create;
        try
          GetKeynames(l);
          CloseKey;

          for n := 0 to l.Count - 1 do
          begin
            CurrentKeyName:=l[n];
            RegBaseName:=RegBaseName+'\'+CurrentKeyName;
            if AnsiLeftStr(RegBaseName,9)<>'\HARDWARE' then exit; {partly matches}
            if RegBaseName<>'\HARDWARE' then {exact match}
              if AnsiContainsStr(RegBaseName,'\ACPI')=False then StopFlag:=True;
            if g_acpiBaseAddress=$CAFE then
            begin
              memo1.Lines.Add('Searching registry key: ' + RegBaseName);
              SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
            end;

            if OpenKey(l[n], False) then
            try
              v := TStringList.Create;
              GetValueNames(v);

              for j:=0 to v.Count -1 do
              begin
                CurrentValueName:=v[j];
                if g_acpiBaseAddress=$CAFE then
                begin
                  memo1.Lines.Add('Reading registry value: ' + CurrentValueName);
                  SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
                end;

                if AnsiContainsStr(RegBaseName,'\FADT\')=True then
                begin
                  s:=GetRegistryData(HKEY_LOCAL_MACHINE,RegBaseName,CurrentValueName);
                  fadt:='';

                  if length(s)>4 then
                  begin
                    for i:=1 to 4 do fadt:=fadt+Chr(Ord(s[i]));
                    memo1.Lines.Add('Found power management registry keyword ' + fadt);
                    SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);

                    {calculate the ACPI register base address. It's a 32-bit pointer but}
                    {I/O port space is limited to 16 bits so we use a 'Word' which is a}
                    {16-bit unsigned integer which fits directly into the DX register}

                    if fadt='FACP' then
                    begin
                      g_acpiBaseAddress:=Ord(s[57]) + (Ord(s[58])*256) + (Ord(s[59])*$10000) + (Ord(s[60])*$1000000);
                      memo1.Lines.Add('ACPI register Base Address = $' + IntToHex(g_acpiBaseAddress,4));
                      SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
                    end;
                  end;
                end;
              end;
            Finally
              v.Free;
            EnumAllKeys(CurrentKey); {recursive, so it calls itself}
            CloseKey;
            k:=LastDelimiter('\',RegBaseName);
            if k>2 then RegBaseName:=AnsiLeftStr(RegBaseName,k-1);
          end;
        end;
      finally
        l.Free
      end;
    finally
      Free;
    end;
end;

{this function returns a variant containing the 'Value' data.}
{this is tricky because of the different data types.}
{invalid or protected values are trapped and simply ignored.}
{if you have trouble reading the values, turn on the 'Raise'}
{and Delphi will tell you why it can't read that value}

function TForm1.GetRegistryData(RootKey: HKEY; Key, Value: string): variant;
var
  Reg: TRegistry;
  RegDataType: TRegDataType;
  DataSize, Len: integer;
  s: string;
label cantread;
begin
  Reg := nil;
  try
    Reg := TRegistry.Create(KEY_QUERY_VALUE);
    Reg.RootKey := RootKey;
    if Reg.OpenKeyReadOnly(Key) then begin
      try
        RegDataType := Reg.GetDataType(Value);
        if (RegDataType = rdString) or
           (RegDataType = rdExpandString) then
          Result := Reg.ReadString(Value)
        else if RegDataType = rdInteger then
          Result := Reg.ReadInteger(Value)
        else if RegDataType = rdBinary then begin
          DataSize := Reg.GetDataSize(Value);
          if DataSize = -1 then goto cantread;
          SetLength(s, DataSize);
          Len := Reg.ReadBinaryData(Value, PChar(s)^, DataSize);
          if Len <> DataSize then goto cantread;
          Result := s;
        end else
cantread:
          // raise Exception.Create(SysErrorMessage(ERROR_CANTREAD));
      except
        s := ''; // Deallocates memory if allocated
        Reg.CloseKey;
        // raise;
      end;
      Reg.CloseKey;
    end else
      // raise Exception.Create(SysErrorMessage(GetLastError));
  except
    Reg.Free;
    // raise;
  end;
  Reg.Free;
end;

{Open the GIVEIO.SYS driver. After opening, the handle is no longer needed.}

Function TForm1.openIO:Boolean;
var
  h:THandle;
begin
  Result:=False;
  Try
    begin
      h:=CreateFile('\\.\\giveio',GENERIC_READ,0,nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
      if h=INVALID_HANDLE_VALUE then exit;
      Result:=True;
      CloseHandle(h); {once we have the driver, we don't need the handle anymore}
    end;
  Finally
  end;
end;

Function TForm1.GetThrottleSlider:Integer;
var SliderPos:Integer;
begin
  SliderPos:=TrackBar1.Position;
  Result:=0;
  Label1.Caption:='Off';
  if SliderPos>0 then
  begin
    if Step625=True then {supports 6.25% step values}
    begin
      Result:=(16-SliderPos)+THROTTLE_ENABLE_BIT;
      Label1.Caption:=SliderText[SliderPos];
    end
    else begin {supports only 12.5% steps}
      Result:=(16-(SliderPos*2))+THROTTLE_ENABLE_BIT;
      Label1.Caption:=SliderText[SliderPos*2];
    end;
  end;
end;

{GIVEIO.SYS provides direct access to the I/O ports}
{so this code also works on Windows NT/2000/2003 and XP}

Procedure TForm1.GetThrottleValFromChip;
var
  ACPIControlReg:Word;
  RegValRead:Byte;

  begin
  ACPIControlReg:= g_acpiBaseAddress + THROTTLE_REGISTER;

  asm
    mov dx,ACPIControlReg
    in al,dx
    and al,$0F //mask off control bits
    mov RegValRead,al
  end;

  {if the readback doesn't work just return 0 to prevent bogus readouts}

  if ReadBackWorks=False then
  begin
    RegValRead:=0;
    memo1.Lines.Add('Readback does''t seem to work, setting value to 0');
    SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
  end;

  memo1.Lines.Add('Current throttle value: $' + IntToStr(RegValRead));
  SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);

  ThrottleValRead:=0;

  if Step625=True then
  begin
    if RegValRead>0 then ThrottleValRead:=16-RegValRead;
  end
  else
  begin
    if RegValRead>0 then ThrottleValRead:=(16-RegValRead) shr 1; {integer divide by 2}
  end;

  TrackBar1.Position:=ThrottleValRead;
  GetThrottleSlider;
end;

function TForm1.DetectChipset:Boolean;
begin
  Result:=False;
  Try
    begin
      EnumAllKEys(HKEY_LOCAL_MACHINE);
      if g_acpiBaseAddress<>$CAFE then Result:=True;
    end;
  Finally
  end;
end;

Procedure TForm1.SetThrottleValue(Throttle:Byte);
var
  ACPIControlReg:Word;
  i:byte;
begin
  if (Throttle=17) and (ReadBackWorks=False) then
  begin
    ShowMessage('It is possible that a value of 93.75% will ' + #13+#10
      + 'set the throttle value back to 0.' + #13 + #10
      + 'In that case, try using a value of 87.5%');
     
  end;
  i:=0;
  if Throttle>0 then i:=Throttle-THROTTLE_ENABLE_BIT;
  memo1.Lines.Add('Setting Throttle Value: $' + IntToStr(i));
  SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
  ACPIControlReg:= g_acpiBaseAddress + THROTTLE_REGISTER;

  asm
    mov dx,ACPIControlReg
    mov al,Throttle
    out dx,al       {set new throttle value}
  end;
  
end;

(*
The throttle probe works by setting a 12.5% throttle value and
reading it back to see if the throttle value catches. After that,
we do the same with a 6.25% throttle. The original throttle
value (that was there before program entry) is restored afterwards.
This probe is executed only once, at program entry. Afterwards,
you no longer have to call it.
*)


Procedure TForm1.ProbeThrottleIncrement;
var
  ACPIControlReg:Word;
  Throttle0:Byte;
  Throttle1:Byte;
  Throttle2:Byte;
  ReadBack:Byte;

begin
  Throttle1:=15+THROTTLE_ENABLE_BIT; {6.25% step}
  Throttle2:=14+THROTTLE_ENABLE_BIT; {12.5% step}
  ACPIControlReg:= g_acpiBaseAddress + THROTTLE_REGISTER;

  memo1.Lines.Add('Probing throttle chip for capabilities.');
  SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);

  asm
    mov dx,ACPIControlReg
    in al,dx
    mov Throttle0,al {backup current throttle value}
    mov al,Throttle2 {first see if it works at 12.5%}
    out dx,al        {set new throttle value}
    and al,$0F       {strip off control bits}
    mov Throttle2,al {and put back into var.}
    in al,dx         {read back & see if it matches}
    and al,$0F       {strip off control bits}
    mov ReadBack,al  {value read back from chip}
  end;

  if Throttle2<>ReadBack then
  begin
      memo1.Lines.Add('Step 12.5%:' + IntToHex(Throttle2,2) + ' Read back:' + IntToHex(ReadBack,2));
      SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
  end
  else
  begin
    ReadBackWorks:=True; {readback appears to work at 12.5% step}
    Memo1.Lines.Add('Register readback for 12.5% steps appears to work.');
    SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
  end;

  {
    This will set the throttle factor to zero upon program entry.
    It could be used to pre-set the throttle factor to some other
    value when the program first starts. Remember that the BIOS often
    allows a 'Thermal Control' setting, which works by pre-charging
    the throttle register with a certain value and activating it
    when needed.
  }
  Throttle0:=0;

  asm
    mov dx,ACPIControlReg
    mov al,Throttle1 {now see if it also works at 6.25%}
    out dx,al        {set new throttle value}
    and al,$0F       {strip off control bits}
    mov Throttle1,al {and save it back to var.}
    in al,dx         {read back & see if it matches}
    and al,$0F       {strip off control bits}
    mov ReadBack,al  {value read back from chip}
    mov al,Throttle0
    out dx,al        {restore original throttle value}
  end;

  if Throttle1<>ReadBack then
  begin
    {Take whatever action you want if you get no 6.25% step}
    memo1.Lines.Add('Step 6.25%. Wrote:$' + IntToHex(Throttle1,2) + ' Read back:$' + IntToHex(ReadBack,2)
      + '. This appears to be a 12.5% step chip.');
    SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
  end
  else
  begin
    {The slider has a default setting of 0..15}
    Step625:=True;
  end;

  {This sets the slider to 0..7}
  if Step625=False then TrackBar1.Max:=7;
end;

{update the (shown, not yet set) percentage whenever the user moves the slider}

procedure TForm1.TrackBar1Change(Sender: TObject);
begin
  GetThrottleSlider;
end;

{
Read the CPU temperature. Googling on the Net, you'll quickly find out how to do it.
For the most common sensor chips (Winbond/Diode) it basically involves selecting
register bank 1 and reading the high and low temp byte. The Winbond chip is usually
connected to a diode, which is the actual temperature sensor. Don't think it will
give you an accurate reading. The readings you get vary from CPU to CPU. It would
be best to calibrate (or compare) the diode's reading with a real thermometer and
add a programmable correction factor (negative or positive offset) to the reading.
}
function TForm1.ReadTemp:Word;
var
  tmp1: byte;
  tmp2: byte;
  Temp: word; // 16-bit unsigned integer
  Temp1: integer;
  i:integer;
begin
  // Wait for the chip to become available. 0x295 is the register which holds
  // the 'busy' flag. Loop while the chip is busy. To prevent endless looping
  // we iterate the polling loop max. 128 times. It shouldn't fail but if it
  // does, just assume that everything is OK and blindly read the values.
  asm
	mov	cx,128    //decimal, poll busy flag max. 128 times
	mov	dx,$0295
@WaitReady1:
	in	al,dx
	and	al,128 //decimal
	jz	@ExitWait
	dec	cx
	jnz	@WaitReady1
@ExitWait:
 	mov	dx,$0295
	mov	al,78 // 78 dec. is the bank select register
	out	dx,al
	inc	dx
	mov	al,1 // select bank 1
	out	dx,al
	mov	dx,0295h
	mov	al,80 // 80 dec. is the temperature High register
	out	dx,al
	inc	dx
	in	al,dx // fetch temp Hi from 81 dec.
	shl	al,1
	mov	tmp1,al
	mov	dx,0295h
	mov	al,81 // 81 dec. is the temperature Low register
	out	dx,al
	inc	dx
	in	al,dx // fetch temp Lo from 82 dec.
	shr	al,7
	mov	tmp2,al
  // calculate reading
	xor	ah,ah
	xor	cx,cx
	mov	al,tmp1
	mov	cl,tmp2
	add	ax,cx
	shr	ax,1
  mov Temp,ax
  end;
{ Calculate average temp over a 5 second period, using a 5-position circular buffer }
  Temp1:=0;
  if (Temp>0) and (Temp<100) then
  begin
    for i:=0 to 3 do AverageTemp[i]:=AverageTemp[i+1];
    AverageTemp[4]:=Temp;
    if AverageTemp[0]=0 then for i:=0 to 3 do AverageTemp[i]:=AverageTemp[4];
    for i:=0 to 4 do Temp1:=Temp1+AverageTemp[i];
    Temp1:=round(Temp1 / 5);
  end;
  if Temp1<0 then Temp1:=0;
  if Temp1>65536 then Temp1:=0;
  Result:=Temp1; // Result is not a variable but holds the function return value
end;

{
This timer goes off every second and reads the CPU temp. The reading is converted
into a variable-length coloured bar. The temp bar is Label2 which sits on top of
a frame (Panel1). This seems the only way to get a properly framed bar. Instead of
a bar you could also use a slider, spin button, or whatever you want. The temp.
readout sub is ReadTemp which returns the temp. In this case, t receives the temp.
Apart from the temp bar, the temp is placed into a circular buffer which is used
to display a temperature histogram. If the histogram buffer fills up, all readings
are moved left 1 position. On the screen, this gives the impression of a
continously moving histogram. The code for this is public domain, author unknown.
}

procedure TForm1.Timer1Timer(Sender: TObject);
var
  n:single;
  i,j,t:integer;
  s:string;
begin
  Timer1.Enabled :=False;
  t:=ReadTemp;
  if (t>0) and (t<100) then
  begin
    // un-comment this if you want a temp-controlled colour bar
    //if t<40 then Label2.Color:=clLime
    //else if t<50 then Label2.Color:=clYellow
    //else
    Label2.Color:=clRed;
    n:=round(Panel1.Width/100) * t;
    s:=FloatToStr(n);
    Val(s,j,i);
    Label2.Width:=j;
    Label2.Caption :=IntToStr(t);
    if TempIdx<Scale then
    begin
      inc(TempIdx);
      TempArray[TempIdx]:=t;
    end
    else
    begin
      for i:=0 to Scale-1 do TempArray[i]:=TempArray[i+1];
      TempArray[Scale-1]:=t;
    end;
    WriteGraph;
    if Checkbox1.Checked then AutoThrottle(t);
    end;
  Timer1.Enabled :=True;
end;

{gray-out spin buttons when not in use}

procedure TForm1.CheckBox1Click(Sender: TObject);
var
  i:Integer;
begin
  if Checkbox1.Checked =True then
  begin
    SpinEdit1.Enabled :=True;
    SpinEdit2.Enabled :=True;
    SpinEdit1.Color :=clWhite;
    SpinEdit2.Color :=clWhite;
    Trackbar1.Enabled:=False;
    Button1.Enabled:=False;
    for i:=0 to 7 do Run[i]:=False;
  end
  else
    begin  //turn off throttle
      TrackBar1.Position:=0;
      SetThrottleValue(GetThrottleSlider);
      GetThrottleValFromChip;
      Trackbar1.Enabled:=True;
      Button1.Enabled:=True;
      SpinEdit1.Enabled :=False;
      SpinEdit2.Enabled :=False;
      SpinEdit1.Color :=clBtnFace;
      SpinEdit2.Color :=clBtnFace;
  end;
end;

{clear the histogram window}

procedure Tform1.ClearImage;
begin
  With Image1.Canvas do begin
    Brush.Color := clWhite;
    FillRect(ClientRect);
  end;
end;

{draw a CPU temperature histogram}

Procedure TForm1.WriteGraph;
Var
  OnePerc: Real;
  MaxX, MaxY : integer;
  DivOne,
  VertDiv : real;
  x, I, IMax, Bottom, Left, Right: Integer;
  S : String;
Begin
  IMax := 10;
  MaxX := Image1.Width - 1;
  MaxY := Image1.Height - 1;
  Bottom := Trunc(MaxY * 0.95);
  Left := Trunc(MaxX * 0.08)-13;
  Right := (Left + Scale+3);
  VertDiv := Bottom / 11;
  DivOne := Bottom / 110;
  OnePerc := 1.0001;
  with Image1.Canvas do begin
    Font.Name := 'Courier';
    Font.Size := 8;
    ClearImage;
    Pen.Color :=clBlack;
    MoveTo(Left,(Bottom - Trunc(VertDiv*10)));
    LineTo(Left,Bottom);
    for x:=1 to 10 do
    begin
      MoveTo(Left, (Bottom-Trunc(VertDiv*x)));
      LineTo(Right, (Bottom-Trunc(VertDiv*x)));
    end;
    MoveTo(Left,Bottom);
    LineTo(Right+1,Bottom);
    MoveTo(Right,(Bottom - Trunc(VertDiv*10)));
    LineTo(Right,Bottom);
    for I := 0 To IMax do begin
      Str(I*10:3,S);
      S := S + '';
      TextOut(Left - 30,(Bottom - Trunc( I * VertDiv) - 7) ,S)
    End;
    For I := 0 To IMax Do Begin
      Str(Trunc(OnePerc * 10 * I),S);
      S := '_' + S;
    End;
    Pen.Color :=clRed;
    For I := 0 To Scale Do begin
      MoveTo(Left+1+I, Bottom-Trunc(TempArray[I]*DivOne/OnePerc));
      LineTo((Left+1+I), Bottom);
    end;
  end;
end;

{
Exit button, same as closing the window.
Triggers Form.Destroy event, which does the cleanup.
}

procedure TForm1.Button2Click(Sender: TObject);
begin
  Application.Terminate;
end;
{
Throttle back one step at a time. First check if the CPU is running
too hot already. If so, we'll probably get very little processor
time to react. So set the throttle to maximum -now-.
Otherwise, gradually throttle back. The hotter the CPU
gets, the more we throttle back. To avoid constant
writing to the chip, we prevent writing the same value
twice in a row. Repeatedly writing to the chip will probably do
no harm but it's a leftover from an older version which shelled
out to DOS to throttle back. It saves a lot of useless actions.
}

procedure TForm1.AutoThrottle(CurrentTemp:Integer);
var
  MaxTemp: Integer;
begin
  MaxTemp:=SpinEdit1.Value;

  if (CurrentTemp>=MaxTemp) and (not Run[7]) then SetAutoThrottle(7);

  if SpinEdit2.Value>0 then
    if (CurrentTemp=(MaxTemp-1)) and (not Run[6]) then SetAutoThrottle(6);
  if SpinEdit2.Value>1 then
    if (CurrentTemp=(MaxTemp-2)) and (not Run[5]) then SetAutoThrottle(5);
  if SpinEdit2.Value>2 then
    if (CurrentTemp=(MaxTemp-3)) and (not Run[4]) then SetAutoThrottle(4);
  if SpinEdit2.Value>3 then
    if (CurrentTemp=(MaxTemp-4)) and (not Run[3]) then SetAutoThrottle(3);
  if SpinEdit2.Value>4 then
    if (CurrentTemp=(MaxTemp-5)) and (not Run[2]) then SetAutoThrottle(2);
  if SpinEdit2.Value>5 then
    if (CurrentTemp=(MaxTemp-6)) and (not Run[1]) then SetAutoThrottle(1);

  if CurrentTemp<(MaxTemp-SpinEdit2.Value) then
  begin
    if Trackbar1.Position>0 then Run[0]:=False; //override any user setting
    if not Run[0] then SetAutoThrottle(0);
  end;
end;

procedure TForm1.SetAutoThrottle(Value:Integer);
var
  i:Integer;
begin
  if Step625 then value:= 2 * value;
  TrackBar1.Position:=Value;
  SetThrottleValue(GetThrottleSlider);
  GetThrottleValFromChip;
  for i:=0 to 7 do Run[i]:=False;
  Run[Value]:=True;
end;

procedure TForm1.MinimizeToTray(Sender: TObject);
begin
  Shell_notifyIcon(NIM_ADD, @TrayIcon);
  Form1.Visible := false;
end;

procedure TForm1.RestoreMainForm;
begin
  Form1.Visible := true;
  Application.BringToFront;
  Application.Restore;
  Shell_notifyIcon(NIM_DELETE, @TrayIcon);
end;

{
The FormDestroy event is triggered when the user terminates the program
in any way, except when they kill it using the Task Manager.
Wait until any pending temp read has finished.
Exit the program only when the temp registers are in a known state,
otherwise the temp readout may freeze up. the only recovery from
messed-up temp regs is a reboot. Upon exit, Throttle is set to 'Off'.
Also, destroy the SysTray icon.
}

procedure TForm1.FormDestroy(Sender: TObject);
begin
  Shell_notifyIcon(NIM_DELETE, @TrayIcon);
  while Timer1.Enabled =False do sleep(100);
  Timer1.Enabled :=False;
  TrackBar1.Position:=0;
  SetThrottleValue(GetThrottleSlider);
end;

{
Tray icon actions
decimal hex symbol effect
513 $201 WM_LBUTTONDOWN left button down
514 $202 WM_LBUTTONUP single left click
515 $203 WM_LBUTTONDBLCLK double left click
516 $204 WM_RBUTTONDOWN right button down
517 $205 WM_RBUTTONUP single right click
518 $206 WM_RBUTTONDBLCLK double right click

If the main form is invisible then all FormMouseMove events are
caught by the SysTray icon.
}

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
begin
  if X = 514 then RestoreMainForm; // single click left
  if X = 517 then // single click right
    if MessageDlg('Exit WinThrottle?',mtConfirmation, mbOKCancel, 0)=mrOK then
      Application.Terminate; // triggers FormDestroy event
end;

procedure TForm1.Restore1Click(Sender: TObject);
begin
  RestoreMainForm;
end;

{ The user closed the form directly, not using the Exit button }
procedure TForm1.Close1Click(Sender: TObject);
begin
  Application.Terminate; // triggers FormDestroy event
end;

procedure TForm1.SetRegistryData(RootKey: HKEY; Key, Value: string;
  RegDataType: TRegDataType; Data: variant);
var
  Reg: TRegistry;
  s: string;
begin
  Reg := TRegistry.Create(KEY_WRITE);
  try
    Reg.RootKey := RootKey;
    if Reg.OpenKey(Key, True) then begin
      try
        if RegDataType = rdUnknown then
          RegDataType := Reg.GetDataType(Value);
        if RegDataType = rdString then
          Reg.WriteString(Value, Data)
        else if RegDataType = rdExpandString then
          Reg.WriteExpandString(Value, Data)
        else if RegDataType = rdInteger then
          Reg.WriteInteger(Value, Data)
        else if RegDataType = rdBinary then begin
          s := Data;
          Reg.WriteBinaryData(Value, PChar(s)^, Length(s));
        end else
          raise Exception.Create(SysErrorMessage(ERROR_CANTWRITE));
      except
        Reg.CloseKey;
        raise;
      end;
      Reg.CloseKey;
    end else
      raise Exception.Create(SysErrorMessage(GetLastError));
  finally
    Reg.Free;
  end;
end;

{
 Save WinThrottle's user setting to HKLM/Software. If all goes well, any user
 should have write permission there.
}

Procedure TForm1.SaveSettings;
var
  s:AnsiString;
begin
  s:='0';
  if Checkbox1.Checked then s:='1';
  SetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0','UseAutoThrottle', rdString, s);
  s:='0';
  if Checkbox2.Checked then s:='1';
  SetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0','StartInSysTray', rdString, s);
  s:=IntToStr(SpinEdit1.Value);
  SetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0','AutoThrottleStart', rdString, s);
  s:=IntToStr(SpinEdit2.Value);
  SetRegistryData(HKEY_LOCAL_MACHINE,'\Software\WinThrottle\Delphi\1.0','ThrottleRange', rdString, s);
end;

// The 'Save' button on the Settings tab.
procedure TForm1.Button3Click(Sender: TObject);
begin
  SaveSettings;
  memo1.Lines.Add('Settings saved to the following registry key:');
  memo1.Lines.Add('HKEY_LOCAL_MACHINE\Software\WinThrottle\Delphi\1.0');
  SendMessage(Memo1.Handle, EM_SCROLLCARET, 0, 0);
  ShowMessage('Settings saved');
end;

procedure TForm1.CheckBox2MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  Memo2.Clear;
  Memo2.Lines.Add('If checked, WinThrottle will startup in the System Tray next time.');
end;

procedure TForm1.CheckBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  Memo2.Clear;
  Memo2.Lines.Add('If checked, WinThrottle will automatically throttle back when the CPU temperature');
  Memo2.Lines.Add('approaches the "Auto-Throttle start" value.');
end;

procedure TForm1.SpinEdit1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  Memo2.Clear;
  Memo2.Lines.Add('This sets the maximum allowed CPU temperature. When this temperature is reached,');
  Memo2.Lines.Add('WinThrottle will set the throttle-back to maximum. Your system will probably become sluggish.');
  Memo2.Lines.Add('');
  Memo2.Lines.Add('This will limit the risk of a CPU burn-out. Of course, whether or not your CPU');
  Memo2.Lines.Add('will actually burn out depends on the capacity of your CPU cooler.');
end;

procedure TForm1.SpinEdit2MouseMove(Sender: TObject; Shift: TShiftState; X,  Y: Integer);
var i1,i2:single;
begin
  Memo2.Clear;
  Memo2.Lines.Add('WinThrottle will gradually start throttling back at ' + IntToStr(SpinEdit2.Value) + ' degrees below the maximum temperature.');
  Memo2.Lines.Add('');
  Memo2.Lines.Add('Current settings are:');
  Memo2.Lines.Add('');
  Memo2.Lines.Add('Throttling starts at ' + IntToStr(SpinEdit1.Value - SpinEdit2.Value) + ' degrees Celsius and will reach maximum throttle at ' + IntToStr(SpinEdit1.Value) + ' degrees Celsius.');
  i1:=Round(((SpinEdit1.Value - SpinEdit2.Value) * 1.8) + 32);
  i2:=Round((SpinEdit1.Value * 1.8) + 32);
  Memo2.Lines.Add('Throttling starts at ' + FloatToStr(i1) + ' degrees Fahrenheit and will reach maximum throttle at ' + FloatToStr(i2) + ' degrees Fahrenheit.');
  Memo2.Lines.Add('');
  Memo2.Lines.Add('If you set this value to zero, WinThrottle will act as a ''safety-valve'' which allows your system to run at full');
  Memo2.Lines.Add('speed until the CPU temperature reaches the maximum value, at which point WinThrottle will abruptly set the throttle to the maximum amount. Your system may appear to ''Freeze up'' until the CPU temperature drops.');
  Memo2.Lines.Add('');
  Memo2.Lines.Add('Set this value above zero to gradually throttle back.');
end;

end.
